/***********************************************************************\
*                               comehere.d                              *
*                                                                       *
*                         Come Here interpreter                         *
*                                                                       *
*                             Version 1.0.1                             *
*                                                                       *
*                            Main entry point                           *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/comehere/.                *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.comehere.comehere;

import smjg.prog.comehere.program;
import std.cstream;
private {
	import std.stdio;
	import std.file;
	import std.path;
	import std.conv;
	import smjg.api.c.disp;
}

version (Windows) {
	extern (C) int isatty(int);
}

int main(char[][] args) {
	try {
		char[] filename;

		switch (args.length) {
			case 2:
				filename = args[1];
				break;

			case 3:
				try {
					if (args[1][0] == '-') {
						processArgs(args[1]);
						filename = args[2];
					} else if (args[2][0] == '-') {
						processArgs(args[2]);
						filename = args[1];
					} else {
						return help();
					}
				} catch (ConvError e) {
					return help();
				}
				break;

			default:
				return help();
		}

		char[] code = cast(char[]) read(defaultExt(filename, "ch"));

		Program program = new Program(code);

		if (TellStatement.useDisp) {
			disp_open();
		}

		program.run();
		return 0;
	} catch(Object e) {
		derr.writefln("Error: %s", e);
		return 1;
	} finally {
		if (TellStatement.useDisp) {
			disp_close();
		}
	}
}


void processArgs(char[] a) {
	version (Windows) {
		if (isatty(stdout._file)) {
			TellStatement.useDisp = true;
		}
	}

	switch (a[1]) {
		case 'a':
			TellStatement.useDisp = false;
			TellStatement.clearData = TellStatement.ANSI;
			break;

		case 'd':
			TellStatement.useDisp = true;
			if (!isatty(stdout._file)) {
				throw new Exception
				  ("Disp API cannot be used when output is redirected");
			}
			break;

		case 'r':
			TellStatement.useDisp = false;
			TellStatement.clearData = \f;
			break;

		default:
			TellStatement.useDisp = false;
			TellStatement.clearData
			  = TellStatement.newLines(toUint(a[1..$]));
	}
}


int help() {
	writefln("comehere: Come Here interpreter, version 1.0.1
Copyright 2007 Stewart Gordon
Usage: comehere [-adr#] filename

Options to control handling of BLANK:
  -a  ANSI escape sequence
  -d  Disp API
  -r  Raw formfeed character
  -#  Number of newlines");
	return 113;
}
