/***********************************************************************\
*                               program.d                               *
*                                                                       *
*                         Come Here interpreter                         *
*                                                                       *
*                             Version 1.0.1                             *
*                                                                       *
*                         The program as a whole                        *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/comehere/.                *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.comehere.program;

import smjg.prog.comehere.parser;
import smjg.prog.comehere.exception;
public import smjg.prog.comehere.statement;
public import smjg.prog.comehere.expression;
private {
	import etc.bigint.bigint_int;
	debug import std.stdio;
}

class Program {
	Statement[] statements;
	Variable[] variables;
	ComeFromStatement[] allComeFroms;
	Int[] labels;

	/**
	 *	Construct a new ComeHereProgram by parsing from an input stream
	 */
	this(char[] code) {
		(new Parser(this)).parse(code);
	}

	/**
	 *	Run the program.
	 */
	void run() {
		foreach (ComeFromStatement cf; allComeFroms) {
			assert (cf !is null);
			debug (trace) {
				writefln("Validating: %s", cf);
			}
			cf.validate();
		}
		debug writefln("Come Froms validated");

		Statement pc = statements[0];

		while (pc !is null) {
			debug (trace) {
				writefln("About to execute: %s", pc);
			}
			pc.execute();
			if (pc.label !is null) {
				ComeFromStatement toGoTo = null;
				foreach (ComeFromStatement cf; allComeFroms) {
					try {
						if (cf.source.value == pc.label) {
							if (toGoTo is null) {
								toGoTo = cf;
							} else {
								throw new ProgramRuntimeException(
								  "Duplicate COME FROM statements "
								  "referencing label " ~ pc.label.toString);
							}
						}
					} catch (UninitialisedVariableException e) {
						/*	Come From statements only take effect from
						 *	when all variables are initialised, so ignore
						 *	any uninits
						 */
					}
				}

				if (toGoTo is null) {
					pc = pc.next;
				} else {
					pc = toGoTo;
				}
			} else {
				pc = pc.next;
			}
		}
	}
}
